"""Python file to run a mammos-mumag simulation."""

import json
import pathlib

import mammos_analysis
import mammos_entity as me
import mammos_mumag
import mammos_units as u
import pandas as pd

HERE = pathlib.Path(__file__).parent.resolve()
u.set_enabled_equivalencies(u.magnetic_flux_field())

with open(HERE / "inp_parameters.json") as f:
    parameters = json.load(f)

H_max = (5 * u.T).to("A/m")

results_hysteresis = mammos_mumag.hysteresis.run(
    mesh="mesh.fly",
    Ms=me.Ms(parameters["Ms"]),
    A=me.A(parameters["A"]),
    K1=me.Ku(parameters["K1"]),
    theta=0,
    phi=0,
    h_start=H_max,
    h_final=-H_max,
    h_n_steps=300,
)

hyst_data = pd.DataFrame(
    {
        "H (A/m)": results_hysteresis.H.q,
        "M (A/m)": results_hysteresis.M.q,
    }
)
hyst_data.to_csv("hystloop.dat", index=False, sep="\t")

# compute extrinsic parameters here and write to disk for convenience. Can also be done
# by `post-processing.ipynb` notebook later.

extrinsic_properties = mammos_analysis.hysteresis.extrinsic_properties(
    results_hysteresis.H,
    results_hysteresis.M,
    demagnetization_coefficient=1 / 3,
)

# Export inputs and outputs to JSON
parameters["Hc"] = extrinsic_properties.Hc.q.to(u.A / u.m).value
parameters["Mr"] = extrinsic_properties.Mr.q.to(u.A / u.m).value
parameters["BHmax"] = extrinsic_properties.BHmax.q.to(u.J / u.m**3).value
with open("parameters.json", "w") as f:
    json.dump(parameters, f, indent=4)



import pathlib
import mochada_kit as mk
from mochada_kit.running import run_plantuml_code

themes_dir = mk._THEMES_DIR

puml_code = f"""@startuml
!theme MOCHADA-CWA from {themes_dir}

:Select a material; <<user_case_input>>
group mammos-spindynamics <<group_single>>
  :Load from SD database; <<data_based_model>>
  :Temperature‐dependent Ms(T); <<raw_output>>
end group

group mammos-analysis <<group_single>>
  :Use Kuzmin model; <<model>>
  :Temperature‐dependent <i>M</i><sub>s</sub>(<i>T</i>), <i>A</i>(<i>T</i>); <<raw_output>>
end group

:Select a working temperature (T); <<user_case_input>>

'  Build & run the micromagnetic model
group optimization <<group_single>>
repeat
    :Propose new geometry; <<user_case_input>>
    group ubermag <<group_single>>
      :Run micromagnetic simulation; <<model>>
      :Output hysteresis loops; <<raw_output>>
    end group
    '  Post‐processing

    group mammos-analysis <<group_single>>
      :Extract linear segment from loops; <<processed_output>>
    end group
repeat while
end group

:Optimal geometry; <<processed_output>>

@enduml
"""

puml_path = pathlib.Path("sensor_workflow.puml")
puml_path.write_text(puml_code, encoding="utf-8")

run_plantuml_code(
    puml_path,
    output_dir=pathlib.Path("."),
)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from __future__ import annotations

from typing import TYPE_CHECKING
from sphinx.domains.python import PythonDomain

from sphinx.ext import intersphinx

if TYPE_CHECKING:
    from sphinx import addnodes, application, environment
    from docutils.nodes import Element
    from sphinx.application import Sphinx
    from sphinx.builders import Builder
    from sphinx.environment import BuildEnvironment
# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "MaMMoS"
copyright = "2025, Thomas Schrefl, Swapneel Amit Pathak, Andrea Petrocchi, Samuel Holt, Martin Lang, Hans Fangohr"
author = "Thomas Schrefl, Swapneel Amit Pathak, Andrea Petrocchi, Samuel Holt, Martin Lang, Hans Fangohr"
# release = '0.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

nitpicky = True
nitpick_ignore = [
    ('py:class', 'owlready2.entity.ThingClass'),
]

extensions = [
    "myst_nb",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]

myst_enable_extensions = [
    "colon_fence",
]
templates_path = ["_templates"]
autosummary_generate = True
autosummary_generate_overwrite = True
autodoc_mock_imports = ["esys-escript"]
autoclass_content = "both"
autodoc_typehints = "description"
autodoc_default_options = {
    # Autodoc members
    "members": True,
    # Autodoc undocumented memebers
    "undoc-members": True,
    # Autodoc private memebers
    "private-members": False,
    # Autodoc special members (for the moment only __init__)
    # "special-members": "__init__",
    "special-members": "__class_getitem__",
}
intersphinx_mapping = {
    "astropy": ("https://docs.astropy.org/en/stable", None),
    "ontopy": ("https://emmo-repo.github.io/EMMOntoPy/latest/", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "owlready2": ("https://owlready2.readthedocs.io/en/latest/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "pydantic": ("https://docs.pydantic.dev/latest", None),
    "python": ("https://docs.python.org/3", None),
    "pyvista": ("https://docs.pyvista.org/", None),
}
exclude_patterns = ["**.ipynb_checkpoints"]
numfig = True
numfig_format = {
    'figure': 'Figure %s',
    'table': 'Table %s',
    'code-block': 'Listing %s',
}
nb_execution_mode = "off"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]
html_logo = "_static/logo.png"
html_favicon = "_static/favicon.png"
html_show_sourcelink = False
html_sourcelink_suffix = ''
html_theme_options = {
    "external_links": [
        {"name": "MaMMoS project", "url": "https://mammos-project.github.io"},
    ],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/MaMMoS-project",
            "icon": "fab fa-github-square",
        },
    ],
    "header_links_before_dropdown": 6,
    "logo": {
        "text": "documentation",
    },
    "secondary_sidebar_items": ["page-toc", "notebook-badges"],
}
html_sidebars = {
    "changelog": [],
    "design": [],
}


# -- Code to fix various Sphinx issues related to type hints -----------------

# code snippet for fixing mapping for pathlib.Path, taken from
# https://github.com/tox-dev/pyproject-api/blob/136e5ded8f65fb157c2e5fee5e8e05de9eefcdd4/docs/conf.py
class PatchedPythonDomain(PythonDomain):
    def resolve_xref(  # noqa: PLR0913,PLR0917
        self,
        env: BuildEnvironment,
        fromdocname: str,
        builder: Builder,
        type: str,  # noqa: A002
        target: str,
        node: resolve_xref,
        contnode: Element,
    ) -> Element:
        # fixup some wrongly resolved mappings
        mapping = {
            "pathlib._local.Path": "pathlib.Path",
        }
        if target in mapping:
            target = node["reftarget"] = mapping[target]
        return super().resolve_xref(
            env, fromdocname, builder, type, target, node, contnode
        )


# A custom handler for :py:data:`numpy.typing.ArrayLike`
# Based on https://github.com/sphinx-doc/sphinx/issues/13308
def missing_reference_handler(
    app: application.Sphinx,
    env: environment.BuildEnvironment,
    node: addnodes.pending_xref,
    contnode,
):
    target = node["reftarget"]
    if "." in target and node["reftype"] == "class":
        # Try again as `obj` so we pick up Unions, TypeVars and other things
        if target.startswith("ophyd_async"):
            # Pick it up from our domain
            domain = env.domains[node["refdomain"]]
            refdoc = node.get("refdoc")
            return domain.resolve_xref(
                env, refdoc, app.builder, "data", target, node, contnode
            )
        else:
            # pass it to intersphinx with the right type
            node["reftype"] = "data"
            return intersphinx.missing_reference(app, env, node, contnode)


def setup(app: application.Sphinx):
    app.connect("missing-reference", missing_reference_handler)
    app.add_domain(PatchedPythonDomain, override=True)



import os
import textwrap
from pathlib import Path

import pytest

from mammos import _fetch_examples as fetch_examples


def test_fetch_file():
    url = fetch_examples.url_from_template(
        repository="mammos", version="0.1.0", directory="examples", file="index.rst"
    )
    index_file = fetch_examples.fetch_file(url)
    assert index_file.startswith(b"mammos\n======")


def test_fetch_file_wrong():
    url = fetch_examples.url_from_template(
        repository="mammos",
        version="0.1.0",
        directory="not-examples",
        file="non-existent",
    )
    with pytest.raises(RuntimeError):
        fetch_examples.fetch_file(url)


def test_parse_index_rst():
    content = textwrap.dedent(
        """\
        Our content
        ===========

        .. toctree::
           :caption: Example

           notebook1
           Advanced usage <notebook2>
           notebook3

        That's all, the next toc will NOT be read.

        .. toctree::
           :caption: Other examples

           This will not be read <other_notebook>
        """
    )
    notebooks = fetch_examples.parse_index_rst(content)
    assert notebooks == ["notebook1.ipynb", "notebook2.ipynb", "notebook3.ipynb"]


def test_parse_index_rst_no_toc():
    content = ""
    assert fetch_examples.parse_index_rst(content) == []

    content = textwrap.dedent(
        """\
        Some file
        =========

        This file does not contain a .. toctree:: directive which we can::

           parse.
        """
    )
    assert fetch_examples.parse_index_rst(content) == []


def test_fech_notebooks_for_mammos(tmp_path):
    fetch_examples.fetch_notebooks_for_repo("mammos", "0.1.0", "examples", tmp_path)
    for notebook in [
        "hard-magnet-material-exploration.ipynb",
        "hard-magnet-tutorial.ipynb",
        "sensor.ipynb",
    ]:
        assert (tmp_path / "mammos" / notebook).exists()


def test_main(tmp_path, capsys):
    os.chdir(tmp_path)
    fetch_examples.main()

    assert (tmp_path / "examples" / "mammos").exists()
    assert (tmp_path / "examples" / "mammos-analysis").exists()

    captured = capsys.readouterr()
    assert captured.err == ""
    assert captured.out.startswith("Downloading examples...")
    assert str(Path("examples/mammos-analysis/quickstart.ipynb")) in captured.out



"""Fetch examples (notebooks) for all mammos subpackages from github."""

import importlib
import re
from pathlib import Path

import requests


def url_from_template(repository: str, version: str, directory: str, file: str) -> str:
    """Create url to fetch single file."""
    return (
        f"https://raw.githubusercontent.com/MaMMoS-project/{repository}"
        f"/refs/tags/{version}/{directory}/{file}"
    )


def fetch_file(url: str) -> bytes:
    """Fetch from url and return content as bytes (encoded as utf-8)."""
    response = requests.get(url, timeout=5)
    if response.status_code != 200:
        msg = f"Fetching '{url}' failed:\n{response.text}"
        raise RuntimeError(msg)
    return response.content


# TODO should we replace this with docutils?
def parse_index_rst(content: str) -> list[str]:
    """Parse index.rst file and find content of first TOC (assuming they are notebooks).

    Expected file structure::

        Any amount of

        content::

           before the toc

        .. toc::
           :caption: Options will be ignored

           file1
           Custom titles are supported <file2>

        Anything after the first toc will be ignored.

    Args:
        content: Content of the index file. Split into lines at newline characters.

    Returns:
        List of notebook names in the form 'name.ipynb'.
    """
    notebooks = []
    toc_start = False
    for line in content.split("\n"):
        if not toc_start and not line.startswith(".. toctree::"):
            continue
        elif line.startswith(".. toctree::"):
            toc_start = True
            continue
        elif re.match(r"\s+:", line) or line == "":  # toctree option or empty line
            continue
        elif not re.match(r"\s+", line):
            # line does not start with whitespace -> end of toctree
            # TODO check for consistent indentation
            break

        # line contains a toc element, options:
        # - '   path/to/document'
        # - '   custom title <path/to/document>'

        # NOTEs:
        # - the special element 'self' is not treated correctly
        #   as we do not expect to find it in any of the files we will parse.
        # - absolute paths are not handled correctly
        if match := re.match(r"\s+.*<(.*)>", line):
            notebooks.append(f"{match.groups()[0]}.ipynb")
        else:
            notebooks.append(f"{line.strip()}.ipynb")

    return notebooks


def fetch_notebooks_for_repo(
    repository: str, version: str, repo_dir: str, base_dir: Path
) -> None:
    """Fetch all notebooks for a given repository and write them to output_dir.

    Notebooks are fetched based on the content of `index.rst` in `remode_dir` based
    on `URL_TEMPLATE`. Notebooks are written to a new sub directory `output_dir/repo`.
    An existing directory with the same name will lead to a failure.

    If no notebooks were found nothing is written to disk.
    """
    index_url = url_from_template(repository, version, repo_dir, "index.rst")
    index_file = fetch_file(index_url)

    notebooks = parse_index_rst(index_file.decode("utf-8"))

    if not notebooks:
        return

    output_dir = base_dir / repository
    output_dir.mkdir()

    for notebook in notebooks:
        notebook_url = url_from_template(repository, version, repo_dir, notebook)
        notebook_content = fetch_file(notebook_url)
        (output_dir / notebook).write_bytes(notebook_content)


def show_files(base_dir: Path) -> None:
    """Print a list of all notebooks in the base_dir subtree."""
    print("The following examples have been downloaded:")
    for notebook in sorted(base_dir.rglob("*.ipynb")):
        print(notebook)


def main():
    """Fetch examples for all repositories.

    At the time of writing this produces the following output::

        examples/
        ├── mammos
        │   ├── hard-magnet-material-exploration.ipynb
        │   ├── hard-magnet-tutorial.ipynb
        │   └── sensor.ipynb
        ├── mammos-analysis
        │   └── quickstart.ipynb
        ├── mammos-dft
        │   └── quickstart.ipynb
        ├── mammos-entity
        │   └── quickstart.ipynb
        ├── mammos-mumag
        │   └── quickstart.ipynb
        ├── mammos-spindynamics
        │   └── quickstart.ipynb
        └── mammos-units
            ├── example.ipynb
            └── quickstart.ipynb

    """
    base_dir = Path("examples")
    base_dir.mkdir()
    print("Downloading examples...")
    for package in [
        "mammos",
        "mammos-analysis",
        "mammos-dft",
        "mammos-entity",
        "mammos-mumag",
        "mammos-spindynamics",
        "mammos-units",
    ]:
        module = importlib.import_module(package.replace("-", "_"))
        fetch_notebooks_for_repo(package, module.__version__, "examples", base_dir)

    show_files(base_dir)



import pathlib
import mochada_kit as mk
from mochada_kit.running import run_plantuml_code

themes_dir = mk._THEMES_DIR

puml_code = f"""@startuml
!theme MOCHADA-CWA from {themes_dir}
split
  -[hidden]->
  split
    -[hidden]->
    :Select a material; <<user_case_input>>
    split
      group mammos-dft <<group_single>>
        :Load from DFT database; <<data_based_model>>
        :Zero temperature <i>M</i><sub>s</sub> & <i>K</i>; <<raw_output>>
      end group
    split again
      group mammos-spindynamics <<group_single>>
        :Load from SD database; <<data_based_model>>
        :Temperature‐dependent <i>M</i><sub>s</sub>(<i>T</i>); <<raw_output>>
      end group
    end split

    group mammos-analysis <<group_single>>
      :Use Kuzmin model; <<model>>
      :Temperature‐dependent <i>M</i><sub>s</sub>(<i>T</i>), <i>K</i>(<i>T</i>), <i>A</i>(<i>T</i>); <<raw_output>>
    end group
  split again
    -[hidden]->
    :Select a working temperature (T); <<user_case_input>>
  end split
  :Evaluate <i>M</i><sub>s</sub>(<i>T</i>), <i>K</i>(<i>T</i>), <i>A</i>(<i>T</i>) at the working temperature; <<output_processing>>
  :<i>M</i><sub>s</sub>, <i>K</i>, <i>A</i> at temperature T; <<processed_output>>
  '  Build & run the micromagnetic model
split again
  -[hidden]->
  :Select a microstructure and other\n micromagnetic input parameters; <<user_case_input>>
split end

group mammos-mumag <<group_single>>
  :Run micromagnetic simulation; <<model>>
  :Hysteresis loop; <<raw_output>>
end group
'  Post‐processing

group mammos-analysis <<group_single>>
  :Extract extrinsic properties; <<data_based_model>>
  :Hc, Mr, BHmax; <<raw_output>>
end group

@enduml
"""

puml_path = pathlib.Path("hard_magnet_workflow.puml")
puml_path.write_text(puml_code, encoding="utf-8")

run_plantuml_code(
    puml_path,
    output_dir=pathlib.Path("."),
)



"""Python file to run all simulations."""

import json
import pathlib
import shutil
import subprocess

import mammos_analysis
import mammos_dft
import mammos_spindynamics
import mammos_units as u
import numpy as np
from mammos_mumag.mesh import Mesh

u.set_enabled_equivalencies(u.magnetic_flux_field())
HERE = pathlib.Path(__file__).parent.resolve()


def setup():
    """Setup hystloop simulation with mumag."""
    OUTDIR = HERE / "out"
    OUTDIR.mkdir(parents=True, exist_ok=True)

    material = "Co2Fe2H4"
    results_dft = mammos_dft.db.get_micromagnetic_properties(material)
    results_spindynamics = mammos_spindynamics.db.get_spontaneous_magnetization(material)
    results_kuzmin = mammos_analysis.kuzmin_properties(
        T=results_spindynamics.T,
        Ms=results_spindynamics.Ms,
        K1_0=results_dft.Ku_0,
    )

    # Download mesh if needed
    if not pathlib.Path("mesh.fly").exists():
        mesh = Mesh("cube50_singlegrain_msize2")
        mesh.write("mesh.fly")

    for T_full in np.linspace(0, 0.95 * results_kuzmin.Tc.value, 16):
        T = round(T_full, 3)  # avoid very large floating point numbers
        print(f"Working on {T=} K.")
        Ms = results_kuzmin.Ms(T)
        A = results_kuzmin.A(T)
        K1 = results_kuzmin.K1(T)

        outdir_i = OUTDIR / f"T_{T}"

        if outdir_i.is_dir():
            # if the output directory is already there, but no parameter file
            # is found, remove the directory and run simulation again
            if (outdir_i / "parameters.json").is_file():
                continue
            else:
                shutil.rmtree(outdir_i)

        outdir_i.mkdir(parents=True)  # raise err if the individual outdir exists
        with open(outdir_i / "inp_parameters.json", "w") as f:
            json.dump(
                {
                    "T": T,
                    "Ms": Ms.value,
                    "A": A.value,
                    "K1": K1.value,
                },
                f,
                indent=4,
            )

        shutil.copyfile(HERE / "submit.sh", outdir_i / "submit.sh")
        shutil.copyfile(HERE / "run.py", outdir_i / "run.py")
        shutil.copy(HERE / "mesh.fly", outdir_i / "mesh.fly")

        res = subprocess.run(
            ["sbatch", "submit.sh"],
            cwd=outdir_i,
            stderr=subprocess.PIPE,
        )

        return_code = res.returncode

        if return_code:
            raise RuntimeError(f"Submission has failed. Exit with error: \n{res.stderr.decode('utf-8')}")

    print("Submission complete.")


if __name__ == "__main__":
    setup()
